#include "../std.h"

#include <string>

#include <globus_gss_assist.h>
#include <globus_io.h>
#include <globus_ftp_control.h>
#include <arpa/inet.h>
#include <sys/socket.h>

#include "userspec.h"
#include "../misc/escaped.h"
#include "../misc/log_time.h"
#include "../misc/inttostring.h"
#include "conf.h"

void userspec_t::free(void) {
//  free(spec->subject);
  if(home)  std::free(home);  home=NULL;
  if(name)  std::free(name);  name=NULL;
  if(group) std::free(group); group=NULL;
//  Groups.clear();
  // Keep authentication info to preserve proxy (just in case)
}

userspec_t::userspec_t(void):name(NULL),group(NULL),home(NULL),gridmap(false) {
}

userspec_t::~userspec_t(void) {
  userspec_t::free();
}

bool userspec_t::unix_user(const char* name_,const char* group_) {
  if(getuid() != 0) return true; // senseless

  struct passwd pw_;
  struct group gr_;
  struct passwd *pw;
  struct group *gr;
  char buf[BUFSIZ];
  char* name_tmp = NULL;
  char* home_tmp = NULL;

  getpwnam_r(name_,&pw_,buf,BUFSIZ,&pw);
  if(pw == NULL) {
    olog<<"Local user "<<name_<<" does not exist"<<endl;
    return false;
  };
  name_tmp=strdup(pw->pw_name);
  home_tmp=strdup(pw->pw_dir);
  if(group_ && group_[0]) {
    getgrnam_r(group_,&gr_,buf,BUFSIZ,&gr);
    if(gr == NULL) {
      olog<<"Local group "<<group_<<" does not exist"<<endl;
      if(name_tmp) std::free(name_tmp); 
      if(home_tmp) std::free(home_tmp);
      return false;
    };
  } else {
    getgrgid_r(pw->pw_gid,&gr_,buf,BUFSIZ,&gr);
    if(gr == NULL) {
      olog<<"User "<<name_<<" has invalid local group id "<<pw->pw_gid<<endl;
      if(name_tmp) std::free(name_tmp); 
      if(home_tmp) std::free(home_tmp);
      return false;
    };
  };
  if(name) std::free(name);
  if(group) std::free(group);
  if(home) std::free(home);
  name=name_tmp;  uid=pw->pw_uid;
  group=strdup(gr->gr_name); gid=gr->gr_gid;
  home=home_tmp;
  olog<<"Remapped to local user: "<<name<<endl;
  olog<<"Remapped to local id: "<<uid<<endl;
  olog<<"Remapped to local group id: "<<gid<<endl;
  olog<<"Remapped to local group name: "<<group<<endl;
  olog<<"Remapped user's home: "<<home<<endl;
  return true;
}

bool check_gridmap(const char* dn,char** user,const char* mapfile) {
  string globus_gridmap;
  if(mapfile) {
    globus_gridmap=mapfile;
  }
  else {
    char* tmp=getenv("GRIDMAP");
    if((tmp == NULL) || (tmp[0] == 0)) {
      globus_gridmap="/etc/grid-security/grid-mapfile";
    }
    else { globus_gridmap=tmp; };
  };
  ifstream f(globus_gridmap.c_str());
  if(!f.is_open() ) {
    olog<<"Mapfile is missing at "<<globus_gridmap<<endl;
    return false;
  };
  for(;!f.eof();) {
    char buf[512]; // must be enough for DN + name
    istream_readline(f,buf,sizeof(buf));
    buf[511]=0;
    char* p = buf;
    for(;*p;p++) if(((*p) != ' ') && ((*p) != '\t')) break;
    if((*p) == '#') continue;
    if((*p) == 0) continue;
    string val;
    int n = input_escaped_string(p,val);
    if(strcmp(val.c_str(),dn) != 0) continue;
    p+=n;
    if(user) {
      n=input_escaped_string(p,val);
      *user=strdup(val.c_str());
    };
    f.close();
    return true;
  };
  f.close();
  return false;
}

int fill_user_spec(userspec_t *spec,globus_ftp_control_auth_info_t *auth,gss_cred_id_t delegated_cred,globus_ftp_control_handle_t *handle) {
  if(spec == NULL) return 1;
  if(!(spec->fill(auth,delegated_cred,handle))) return 1;
  return 0;
}

bool userspec_t::fill(globus_ftp_control_auth_info_t *auth,gss_cred_id_t delegated_cred,globus_ftp_control_handle_t *handle) {
  struct passwd pw_;
  struct group gr_;
  struct passwd *pw;
  struct group *gr;
  char buf[BUFSIZ];

  if(auth->auth_gssapi_subject == NULL) return false;
  string subject = auth->auth_gssapi_subject;
  make_unescaped_string(subject);
  name=NULL;
  if(!check_gridmap(subject.c_str(),&name)) {
    olog<<"Warning: there is no local mapping for user"<<endl;
  } else {
    if((name == NULL) || (name[0] == 0)) {
      olog<<"Warning: there is no local name for user"<<endl;
    } else {
      gridmap=true;
    };
  };
  // fill host info
  if(globus_io_tcp_get_remote_address(&(handle->cc_handle.io_handle),
         host,&(port)) != GLOBUS_SUCCESS) {
    port=0;
    user.set(auth->auth_gssapi_subject,delegated_cred);
  } else {
    char abuf[1024]; abuf[sizeof(abuf)-1]=0;
    //int host[4] = {0,0,0,0};
    //unsigned short port = 0;
    struct hostent he;
    struct hostent* he_p;
    struct in_addr a;
    snprintf(abuf,sizeof(abuf)-1,"%u.%u.%u.%u",host[0],host[1],host[2],host[3]);
    if(inet_aton(abuf,&a) != 0) {
      int h_errnop;
      char buf[1024];
      he_p=globus_libc_gethostbyaddr_r((char*)&a,strlen(abuf),AF_INET,
                                             &he,buf,sizeof(buf),&h_errnop);
      if(he_p) {
        if(strcmp(he_p->h_name,"localhost") == 0) {
          abuf[sizeof(abuf)-1]=0;
          if(globus_libc_gethostname(abuf,sizeof(abuf)-1) != 0) {
            strcpy(abuf,"localhost");
          };
        };
      };
    };
    user.set(auth->auth_gssapi_subject,delegated_cred,abuf);
  };
  if((user.proxy() == NULL) || (user.proxy()[0] == 0)) {
    olog<<"No proxy provided"<<endl;
  } else {
    olog<<"Proxy stored at "<<user.proxy()<<endl;
  };
  if((getuid() == 0) && name && name[0]) {
    olog<<"Mapped to local user: "<<name<<endl;
    getpwnam_r(name,&pw_,buf,BUFSIZ,&pw);
    if(pw == NULL) {
      olog<<"Local user does not exist"<<endl;
//      ::free(subject); subject=NULL;
      ::free(name); name=NULL;
      return false;
    };
  }
  else {
    if(name) ::free(name); name=NULL;
    getpwuid_r(getuid(),&pw_,buf,BUFSIZ,&pw);
    if(pw == NULL) {
      olog<<"Warning: running user has no name"<<endl;
//      subject=strdup("unknown");
    }
    else {
      name=strdup(pw->pw_name);
      olog<<"Mapped to running user: "<<name<<endl;
    };
  };
  if(!(name)) name=strdup("");
  uid=pw->pw_uid;
  gid=pw->pw_gid;
  olog<<"Mapped to local id: "<<uid<<endl;
  olog<<"Mapped to local group id: "<<gid<<endl;
  home=strdup(pw->pw_dir);
  olog<<"Mapped user's home: "<<home<<endl;
  getgrgid_r(gid,&gr_,buf,BUFSIZ,&gr);
  if(gr == NULL) {
    olog<<"Invalid local group"<<endl;
//    if(subject) ::free(subject); subject=NULL;
    if(name) ::free(name); name=NULL;
    if(home) ::free(home); home=NULL;
    return false;
  };
  group=strdup(gr->gr_name);
  olog<<"Mapped to local group name: "<<group<<endl;
//  delegated_cred=delegated_cred;
  return true;
}

bool userspec_t::fill(AuthUser& u) {
  struct passwd pw_;
  struct group gr_;
  struct passwd *pw;
  struct group *gr;
  char buf[BUFSIZ];

  string subject = u.DN();
  name=NULL;
  if(!check_gridmap(subject.c_str(),&name)) {
    olog<<"Warning: there is no local mapping for user"<<endl;
  } else {
    if((name == NULL) || (name[0] == 0)) {
      olog<<"Warning: there is no local name for user"<<endl;
    } else {
      gridmap=true;
    };
  };
  user=u;
  if((user.proxy() == NULL) || (user.proxy()[0] == 0)) {
    olog<<"No proxy provided"<<endl;
  } else {
    olog<<"Proxy stored at "<<user.proxy()<<endl;
  };
  if((getuid() == 0) && name && name[0]) {
    olog<<"Mapped to local user: "<<name<<endl;
    getpwnam_r(name,&pw_,buf,BUFSIZ,&pw);
    if(pw == NULL) {
      olog<<"Local user does not exist"<<endl;
      ::free(name); name=NULL;
      return false;
    };
  } else {
    if(name) ::free(name); name=NULL;
    getpwuid_r(getuid(),&pw_,buf,BUFSIZ,&pw);
    if(pw == NULL) {
      olog<<"Warning: running user has no name"<<endl;
//      subject=strdup("unknown");
    }
    else {
      name=strdup(pw->pw_name);
      olog<<"Mapped to running user: "<<name<<endl;
    };
  };
  if(!(name)) name=strdup("");
  uid=pw->pw_uid;
  gid=pw->pw_gid;
  olog<<"Mapped to local id: "<<uid<<endl;
  olog<<"Mapped to local group id: "<<gid<<endl;
  home=strdup(pw->pw_dir);
  olog<<"Mapped user's home: "<<home<<endl;
  getgrgid_r(gid,&gr_,buf,BUFSIZ,&gr);
  if(gr == NULL) {
    olog<<"Invalid local group"<<endl;
    if(name) ::free(name); name=NULL;
    if(home) ::free(home); home=NULL;
    return false;
  };
  group=strdup(gr->gr_name);
  olog<<"Mapped to local group name: "<<group<<endl;
  //// fill host info here
  //if(globus_io_tcp_get_remote_address(&(handle->cc_handle.io_handle),
  //       host,&(port)) != GLOBUS_SUCCESS) {
  //  port=0;
  //};
  return true;
}

string subst_user_spec(string &in,userspec_t *spec) {
  string out = "";
  int i;
  int last;
  last=0;
  for(i=0;i<in.length();i++) {
    if(in[i] == '%') {
      if(i>last) out+=in.substr(last,i-last);
      i++; if(i>=in.length()) {         };
      switch(in[i]) {
        case 'u': {
          char buf[10];
          snprintf(buf,9,"%i",spec->uid); out+=buf; last=i+1; 
        }; break;
        case 'U': { out+=spec->name; last=i+1; }; break;
        case 'g': {
          char buf[10];
          snprintf(buf,9,"%i",spec->gid); out+=buf; last=i+1; 
        }; break;
        case 'G': { out+=spec->group; last=i+1; }; break;
        case 'D': { out+=spec->user.DN(); last=i+1; }; break;
        case 'H': { out+=spec->home; last=i+1; }; break;
        case '%': { out+='%'; last=i+1; }; break;
        default: {
          olog << "Warning: undefined control sequence: %" << in[i] << endl;
        };
      };
    };
  };
  if(i>last) out+=in.substr(last);
  return out;
}
